\ (c) 1989 Johns Hopkins University / Applied Physics Laboratory
\ Martin Fraeman 	Apr 28, 1989

\ The words below are already defined in our SC32 kernel.  To run
\ this code on other machines, you made need to also include these 
\ definitions
\ : 2pick 2 pick ;
: <= > not ; 
: >= < not ;
: 2under	( a b c -- c b ) 
	swap rot drop ;
: 3under	( a b c d -- d b c )
	>r rot drop r> -rot ;	
: rover r> r> r> 2dup >r >r nip swap >r ;
: 2rpick r> r> r> r> 2pick 2pick 2pick >r >r >r nip nip swap >r ;
: 3rpick r> r> r> r> r> 3pick 3pick 3pick 3pick 
	>r >r >r >r nip nip nip swap >r ;


\ These system dependent words can time the benchmarks
\ program timing words for the APL SC32 puppy board
decimal
variable tickcnt 	0 tickcnt !		( elapsed number of ticks )
variable tickrate	10 tickrate !	( ticks/sec )
variable repcount	10 repcount !	( repetitions/routine )

\ type the time in tickcnt
: timetype
	base @ decimal
	tickcnt @ tickrate @ repcount @ * /mod 0 
	<# #s #> type ." ." 2 u.lz 
	base !
;

variable tickstart 	0 tickstart !
\ use a fake hardware clock for non-puppy systems
variable hardwareclock	0 hardwareclock !

: timeon 0 tickcnt ! hardwareclock @ tickstart ! ;
: timeoff hardwareclock @ tickstart @ -  tickcnt +! ;
: timepause hardwareclock @ tickstart @ - tickcnt +! ;
: timeresume hardwareclock @ tickstart ! ;


\ These structure access words were originally developed by 
\ at JHU/APL by Ben Ballard and John Hayes
\ Structure access words
\ Examples of use:
\	structure foo		\ declare a structure named foo
\	      wrd: .thing1	\ with a one word field named .thing1
\	   2 wrds: .thing2	\ and a two word field named .thing2
\	endstructure
\	
\	structure foobar	\ another structure
\	          wrd: .thing
\	   foo struct: .blah	\ nested structure
\	endstructure
\
\	foobar makestruct test	\ allocate space for a structure instance
\	1234 test .blah .thing1 ! \ access structure

: structure	\ ( --- here 0 ) structure access
   create
      here 0 , 0
   does> @ ;	\ ( --- size )

: struct:	\ ( offset size --- offset' ) nested structure definition
   create
      over , +
   does> @ + ;	\ ( addr --- addr' ) add member's offset to addr.

: wrds:		\ ( offset size --- offset' ) create structure member size
		\ words long starting at offset.
   cells struct: ;

: wrd:		\ ( offset --- offset' ) create structure member 1 word long
		\ starting at offset.
   cell struct: ;

: endstructure	\ ( structdef size --- )
   swap ! ;

: makestruct	\ ( size --- ) allocate memory for a struct of given size
   variable cell- allot ;
decimal


\ This is a suit of benchmarks that are relatively short, both in program
\ size and execution time.  It requires no input, and prints the execution
\ time for each program, using the system-dependent words timeon, 
\ timeoff, and timetype, given below for APL's SC32 prototype board 
\ to measuer the CPU time for each program.  It does a rudimentary check to
\ make sure each program gets the right output.  These programs were
\ gathered by John Hennessy and modified by Peter Nye.  Martin Fraeman at 
\ JHU/APL converted the programs from C to Forth.



\ random number generator

variable seed		\ int seed

: Initrand	( -- )	\ initialize random number generator
	74755 seed !
;

: Rand		( -- n)	\ generate a random number
	seed @ 1309 * 13849 + 65535 and 
	dup seed !
;



\ Permutation program, heavily recursive, written by Denny Brown
10 constant permrange						\ #define permrange 10
create permarray permrange 1+ cells allot	\ int permarray[permrange+1]
variable pctr								\ int pctr

: Swap	( a b -- )		\ swap the contents of memory locations a and b
	dup @ >r 			( a b - a b ) ( -- b@ )
	over @				( a b -- a b a@ ) ( -- b@ )
	over ! drop			( a b a@ -- a ) ( b@ -- b@ )
	r> over ! drop		( a -- ) ( b@ -- )
;


: Initialize	( -- )	\ initialze the permutation array
	8 1 do
		i 1-
		permarray i cells + !
	loop
;

: Permute	( n -- )	\ scramble the array
recursive
	1 pctr +!
	dup 1 = not
	if
		dup 1-						( n -- n n-1 )
		dup Permute					( n n-1 -- n k )
		begin
			dup 0 >
		while
			over cells permarray + 	\ &permarray[n]
			over cells permarray +	\ &permarray[k]
			Swap
			over 1- Permute
			over cells permarray + 	\ &permarray[n]
			over cells permarray +	\ &permarray[k]
			Swap
			1-
		repeat drop
	then drop
;

: Perm	( -- )		\ Execute the permutations test
	0 pctr !
	1
	begin
		dup 6 <
	while
		Initialize
		7 Permute
		1+
	repeat drop
	pctr @ 43300 = not
	if ." Error in Perm" cr then
;


\ Program to Solve the Towers of Hanoi
structure element
	wrd: .discsize
	wrd: .next
endstructure

: element* 2* ;

18 constant maxcells
3 constant stackrange
create stack 		stackrange 1+ cells allot
create cellspace 	maxcells 1+ element* cells allot

variable freelist
variable movesdone


: Makenull	( s -- )		\ clean out a stack disk
    0 over cells stack + ! drop
;

: Getelement	( -- s )	\ get a new stack element index
				\ maintain temp on tos
	freelist @ 0>
	if
		freelist @
		dup element* cells cellspace + .next @ freelist !
	else
		." Tower: out of space" cr 
		0		\ temp not set but returned in original 
	then
;


: Push			( i s -- )	\ push disk i onto stack s
	false >r				 \ errorfound on tor, localel on tos
	dup cells stack + @ 0>		
	if
		dup cells stack + @ 
		element* cells cellspace + .discsize @ 2pick <=
		if 
			rdrop true >r		\ errorfound = true
			." Tower: disc size error" cr 
		then
	then
	r> not
	if
		Getelement									( i s -- i s l )
		over cells stack + @ 
		over element* cells cellspace + .next !		( i s l -- i s l )
		dup 2pick cells stack + !					( i s l -- i s l )
		element* cells cellspace + .discsize nip !	( i s l -- )
	then
;

: Init				( s n -- )	\ Initialize tower s with n discs
	over Makenull
	begin
		dup 1 >=
	while
		dup 2pick Push
		1-
	repeat
	2drop 
;

: Pop				( s -- n )	\ Pop tower s, return new tower size
	dup cells stack + @		( s -- s stack[s] )
	dup 0>
	if
		dup element* cells cellspace + 
		.discsize @ >r 					( s s[s] -- s s[s] ) ( -- temp1 )
		dup element* cells cellspace + 
		.next @ >r 						( s s[s] -- s s[s] ) ( -- temp1 temp )
		freelist @ over element* cells cellspace + .next !
		freelist !
		r> over cells stack + !
		drop r>
	else
		." Tower: nothing to pop " cr
		drop
	then
;

: Move					( s1 s2 -- )	\ move a disc from s1 to s2
	>r Pop r> Push
	1 movesdone +!
;

: tower recursive		( i j k -- )	\ source dest #discs
	dup 1 =
	if
		drop Move
	else
		6 3pick - 2pick - >r	( i j k -- i j k ) ( -- other )
		1- 2pick r@ 2pick tower
		2pick 2pick Move
		r> 2pick 2pick tower
		2drop drop
	then
;

: Towers	( -- )		\ Towers of Hanoi with 14 discs
	maxcells 1+ 1
	do
		i element* cells cellspace + 
		i 1- over .next !
		0 over .discsize ! drop
	loop
 	maxcells freelist !
 	1 14 Init
 	2 Makenull
 	3 Makenull
 	0 movesdone !
 	1 2 14 tower
 	movesdone @ 16383 = not
 	if
 		." Error in Towers." cr
 	then
;


\ The eight queens problem

\ The original C was written assuming that 6 arguments
\ were passed to the Try routine.  These were all local
\ to Doit and were never changed in subsequent invocations
\ of Try.  At first I used local variables to handle this
\ mess (they worked fine by the way) but since there seems
\ to no concensus on locals should work among Forth implementations
\ I've rehacked this version to assume the arguments are
\ available globally.  That probably violates the spirit of
\ the original benchmark a little more than usual however.

decimal

create x 	9 1+ cells allot
create c	15 1+ cells allot
create b	17 1+ cells allot
create a	9 1+ cells allot
variable q

: Try	( i -- )
recursive
	0			\ keep j on top of stack
	false q !
	begin
		q @ not over 8 <> and
	while							( -- i j )
		1+
		false q !
		dup cells a + @ 
		2pick 2pick + cells b + @ and
		2pick 2pick - 7 + cells c + @ and 
		if
			dup 2pick cells x + !
			false over cells a + !
			false 2pick 2pick + cells b + !
			false 2pick 2pick - 7 + cells c + !
			over 8 <
			if
				over 1+ Try
				q @ not
				if
					true over cells a + !
					true 2pick 2pick + cells b + !
					true 2pick 2pick - 7 + cells c + !
				then
			else true q !
			then
		then
	repeat 2drop
;

: Doit		( -- )		\ Make the queens happy
	-7					\ i=0-7
	begin
		dup 16 <=
	while
		dup 1 >= over 8 <= and
		if true over cells a + ! then
		dup 2 >=
		if true over cells b + ! then
		dup 7 <=
		if true over 7 + cells c + ! then
		1+
	repeat drop
	1 Try 
	q @ not
	if
		." Error in Queens" cr
	then
;

: Queens
	50 0 do Doit loop
;


\ Multiplies two integer matrices.

decimal
40 constant rowsize

\ quick multiply by 40
: rowsize* 2* 2* 2* dup 2* 2* + ;

create ima rowsize rowsize* cells allot
create imb rowsize rowsize* cells allot
create imr rowsize rowsize* cells allot


: Initmatrix	( m -- )	\ int m[rowsize+1][rowsize+1]
	rowsize 0 do
		rowsize 0 do
			dup Rand
			dup 120 / 120 * 60 - -
			swap j rowsize* i + cells + !
		loop
	loop
	drop
;

: Innerproduct	( a b row col -- result )
\ compute the inner product of A[row,*] and B[*,column]
	>r >r 0									( a b r c -- a b y)  ( -- c r)
	1 >r									( -- a b y) ( -- c r i)
	begin
		r@ rowsize <=
	while									( -- a b y) ( -- c r i)
		2pick rover rowsize* r@ + cells + @ 	( -- a b y a[r][i]) 
		2pick r@ rowsize* 2rpick + cells + @	( -- a b y a[r][i] b[i][c])
		* +									( -- a b y)
		r> 1+ >r							( -- a b y) ( -- c r i)
	repeat
	rdrop rdrop rdrop nip nip
;
		

: Intmm		( -- )
	Initrand ima Initmatrix imb Initmatrix
	rowsize 0 do
		rowsize 0 do
			ima imb j i Innerproduct
			i rowsize* j + cells imr + !
		loop
	loop
;


\ The infamous mysterious compute bound puzzle attributed to Forest Baskett
decimal

3 constant classmax
12 constant typemax
511 constant size
8 constant D

create piececount	classmax 1+ cells allot
create class		typemax 1+ cells allot
create piecemax		typemax 1+ cells allot
create puzzl		size 1+ cells allot
create p			typemax 1+ size 1+ * 1+ cells allot

variable kount

: size* dup >r 2* 2* 2* 2* 2* 2* 2* 2* 2* r> - ;

: Fit		( i j -- flag )
	over cells piecemax + @ 1+ 0 
	do										\ for(k=0;k<=piecemax[i];k++)
		over size* i + cells p + @		\ if(p[i][k])
		if dup i + cells puzzl + @			\ if(puzzl[j+k])
			if 2drop false undo exit then	\ return(false)
		then
	loop 
	2drop true
;

: Place		( i j -- flag )
	over cells piecemax + @ 1+ 0
	do										\ for(k=0;k<=piecemax[i];k++)
		over size* i + cells p + @		\ if p[i][k]
		if 
			true over i + cells puzzl + ! 	\ puzzl[j+k]=true
		then	
	loop
	over cells class + @ cells piececount +	\ &piececount[class[i]]
	dup >r @ 1- r> !						\ piececount[class[i]]--
	size 1+ over 
	do										\ for(k=j;k<=size;k++)
		i cells puzzl + @ not				\ if(!puzzl[k])
		if 2drop i undo exit then			\ return(k)
	loop
	2drop 0									\ return(0)
;

: Remove	( i j -- )
	over cells piecemax + @ 1+ 0 
	do										\ for(k=0;k<=piecemax[i];k++)
		over size* i + cells p + @		\ if(p[i][k])
		if 
			false over i + cells puzzl + !	\ puzzl[j+k]=false
		then
	loop
	over cells class + @ cells piececount +	\ &piececount[class[i]]
	dup >r @ 1+ r> ! 2drop					\ piececount[class[i]]++
;

: Trial		( j -- flag )
recursive
	1 kount +!								\ kount=kount+1
	typemax 1+ 0
	do										\ for(i=0;i<=typemax;i++)
		i cells class + @ cells piececount + @ 0 <>
		if									\ if(piececount[class[i]])
			i over Fit						\ if(Fit(i,j)
			if
				i over Place				\ tos=k=Place(i,j)
				dup Trial swap 0= or 		\ if(Trial(k)||(k==0)
				if
					drop true undo exit		\ return(true)
				else
					i over Remove 			\ Remove(i,j)
				then
			then
		then
	loop
	drop false
;


: Puzzle		( -- )						\ the real puzzle
	size 1+ 0 do true i cells puzzl + ! loop	\ puzzl[]=true
	6 1 do i 6 1 do 6 1 do					\ (i=1:5)(j=1:5)(k=1:5)
				false
				D i * j + D * 2pick + cells	\ [i+d*(j+d*k)]
				puzzl + !
	loop loop drop loop
	typemax 1+ 0 do size 1+ 0 do 
			false j size* i + cells p + !	\ p[i][m]=false
	loop loop
	4 0 do i 2 0 do 1 0 do					\ (i=0:3)(j=0:1)(k=0:0)
				true
				D i * j + D * 2pick + cells	\ [i+d*(j+d*k)]
				p + !
	loop loop drop loop
	0 class !
	3 D + piecemax !
	2 0 do i 1 0 do 4 0 do					\ (i=0:1)(j=0:0)(k=0:3)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				size + cells p + !
	loop loop drop loop
	0 1 cells class + !
	1 D D * 3 * + 1 cells piecemax + !
	1 0 do i 4 0 do 2 0 do					\ (i=0:0)(j=0:3)(k=0:1)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				2 size* + cells p + !
	loop loop drop loop
	0 2 cells class + !
	D 3 * D D * + 2 cells piecemax + !
	2 0 do i 4 0 do 1 0 do					\ (i=0:1)(j=0:3)(k=0:0)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				3 size* + cells p + !
	loop loop drop loop
	0 3 cells class + !
	1 D 3 * + 3 cells piecemax + !
	4 0 do i 1 0 do 2 0 do					\ (i=0:3)(j=0:0)(k=0:1)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				4 size* + cells p + !
	loop loop drop loop
	0 4 cells class + !
	3 D D * + 4 cells piecemax + !
	1 0 do i 2 0 do 4 0 do					\ (i=0:0)(j=0:1)(k=0:3)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				5 size* + cells p + !
	loop loop drop loop
	0 5 cells class + !
	D D D * 3 * + 5 cells piecemax + !
	3 0 do i 1 0 do 1 0 do					\ (i=0:2)(j=0:0)(k=0:0)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				6 size* + cells p + !
	loop loop drop loop
	1 6 cells class + !
	2 6 cells piecemax + !
	1 0 do i 3 0 do 1 0 do					\ (i=0:0)(j=0:2)(k=0:0)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				7 size* + cells p + !
	loop loop drop loop
	1 7 cells class + !
	D 2 * 7 cells piecemax + !
	1 0 do i 1 0 do 3 0 do					\ (i=0:0)(j=0:0)(k=0:3)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				8 size* + cells p + !
	loop loop drop loop
	1 8 cells class + !
	D D * 2 * 8 cells piecemax + !
	2 0 do i 2 0 do 1 0 do					\ (i=0:1)(j=0:1)(k=0:0)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				9 size* + cells p + !
	loop loop drop loop
	2 9 cells class + !
	1 D + 9 cells piecemax + !
	2 0 do i 1 0 do 2 0 do					\ (i=0:1)(j=0:0)(k=0:1)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				10 size* + cells p + !
	loop loop drop loop
	2 10 cells class + !
	1 D D * + 10 cells piecemax + !
	1 0 do i 2 0 do 2 0 do					\ (i=0:0)(j=0:1)(k=0:1)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				11 size* + cells p + !
	loop loop drop loop
	2 11 cells class + !
	D D D * + 11 cells piecemax + !
	2 0 do i 2 0 do 2 0 do					\ (i=0:1)(j=0:1)(k=0:1)
				true
				D i * j + D * 2pick + 		\ [i+d*(j+d*k)]
				12 size* + cells p + !
	loop loop drop loop
	3 12 cells class + !
	1 D + D D * + 12 cells piecemax + !

	13 piececount !							\ piececount[0]=13
	3 1 cells piececount + !				\ piececount[1]=3
	1 2 cells piececount + !				\ piececount[2]=1
	1 3 cells piececount + !				\ piececount[3]=1
	D 1+ D * 1+								\ tos=m=1+d*(1+d*1)
	0 kount !

	0 over Fit								\ if(Fit(0,m)
	if 0 swap Place 						\ n=Place(0,m)
	else ." Error 1 in Puzzle" cr
	then
	Trial not								\ if(!Trial(n))
	if ." Error2 in Puzzle" cr 
	else 2005 kount @ <> if ." Error3 in Puzzle" cr then
	then
;


\ Sorts an array using quicksort
decimal

variable biggest
variable littlest
variable top

5000 constant sortelements
500 constant srtelements

create sortlist 	sortelements 1+ cells allot

: Initarr	( -- )
	Initrand 0 biggest ! 0 littlest !
	sortelements 1+ 1
	do
		Rand dup 100000 / 100000 * - 50000 - 
		dup i cells sortlist + !
		dup biggest @ >
		if biggest !
		else dup littlest @ <
		if littlest ! else drop then
		then
	loop
;

: Quicksort	( a l r -- )
recursive	( naturally)
	dup >r over >r				( a l r -- a i j ) ( -- r l )
	2dup + 2/ cells 3pick + @	( a i j -- a i j x )
	begin
		begin
			3pick 3pick cells + @ over <	\ a[i]<x
		while
			2pick 1+ 3under					\ i=i+1
		repeat
		begin
			3pick 2pick cells + @ over > 	\ x<a[j]
		while
			over 1- 2under					\ j=j-1
		repeat
		2pick 2pick <= 						\ if(i<=j)
		if 									\ xchng a[i],a[j]
			>r							( a i j x -- a i j) ( r l -- r l x )
			2pick dup 3pick cells +	dup @ ( a i j -- a i j a ai a[i])
			3pick cells 3pick + dup @	( -- a i j a ai a[i] aj a[j] )
			3pick ! ! 2drop				( -- a i j )
			over 1+ 2under 1-				\ i=i+1; j=j-1
			r>							( a i j -- a i j x) ( r l x -- r l)
		then
			2pick 2pick > 	 				\ while(i<=j)
 	until
	drop								( a i j x -- a i j)
	r@ over < 								\ if(l<j)
	if 2pick r@ 2pick Quicksort then		\ Quicksort(a,l,j)
	rdrop drop r> 2dup <					\ if(i<r)
	if 2pick 2pick 2pick Quicksort then		\ Quicksort(a,i,r)
	2drop drop
;

: Quick
	Initarr sortlist 1 sortelements Quicksort
	sortlist cell+ @ littlest @ <>
	sortelements cells sortlist + @ biggest @ <> or
	if ." Error in Quick" cr then
;



\ Perform a tree sort
0 constant nil

structure node
	wrd: .left
	wrd: .right
	wrd: .val
endstructure

node makestruct tree


: malloc		( n -- a )		\ allot n cells to hold a nameless struct
	>r here r> allot
;

: CreateNode	( t n -- )
	node malloc 2dup .val ! nip			( t n -- t s )
	nil over .left !
	nil over .right !
	over ! drop
;

: Insert		( n t -- )		\ insert n into tree
recursive
	2dup .val @ >			\ if(n>t->val)
	if dup .left @ nil =			\ if(t->left==nil)
		if dup .left 2pick CreateNode
		else 2dup .left @ Insert
		then
	else 2dup .val @ <
		if dup .right @ nil =
			if dup .right 2pick CreateNode
			else 2dup .right @ Insert
			then
		then
	then
	2drop
;


: Checktree		( p -- f )	\ check by inorder traversal
recursive
	true >r
	dup .left @ nil <>
	if
		dup .left @ .val @ over .val @ <=
		if rdrop false >r 
		else dup .left @ Checktree r> and >r
		then
	then 
	dup .right @ nil <>
	if
		dup .right @ .val @ over .val @ >=
		if rdrop false >r
		else dup .right @ Checktree r> and >r
		then
	then
	drop r>
;

: Trees			( -- )
	here
	Initarr node malloc
	nil over .left !
	nil over .right !
	sortlist cell+ @ over .val !
	sortelements 1+ 2 
	do sortlist i cells + @ over Insert loop
	Checktree not
	if ." Error in Tree" cr then
	dp !
;
		

\ Finally, the silly bubble sort
: bInitarr	( -- )
	Initrand 0 biggest ! 0 littlest !
	srtelements 1+ 1
	do
		Rand dup 100000 / 100000 * - 50000 - 
		dup i cells sortlist + !
		dup biggest @ >
		if biggest !
		else dup littlest @ <
		if littlest ! else drop then
		then
	loop
;

: Bubble
	bInitarr srtelements 
	begin
		dup 1 >
	while
		1
		begin
			dup 2pick <					( t i -- t i f )
		while
			dup cells sortlist + 		( -- t i si )
			over 1+ cells sortlist + 	( -- t i si si+ )
			over @ over @ >	
			if							( -- t i si si+ )
				dup @ 2pick @			( -- t i si si+ s[i+] s[i] )
				2pick ! 2pick !			( -- t i si si+ )
			then
			2drop 1+
		repeat drop
		1-
	repeat drop
	sortlist cell+ @ littlest @ <>
	sortlist srtelements cells + @ biggest @ <> or
	if
		." Error3 in Bubble" cr
	then
;

: TIME	( f -- )	\ time 10 executions of function f
	timeon
	repcount @ 0 do
		dup timeresume execute timepause
	loop drop 
;
		

: Stan
cr
4 spaces ." Perm " 4 spaces 	['] Perm TIME timetype cr
2 spaces ." Towers " 4 spaces 	['] Towers TIME timetype cr
2 spaces ." Queens " 4 spaces 	['] Queens TIME timetype cr
3 spaces ." Intmm " 4 spaces 	['] Intmm TIME timetype cr
2 spaces ." Puzzle " 4 spaces 	['] Puzzle TIME timetype cr
3 spaces ." Quick " 4 spaces 	['] Quick TIME timetype cr
2 spaces ." Bubble " 4 spaces 	['] Bubble TIME timetype cr
4 spaces ." Tree " 4 spaces 	['] Trees TIME timetype cr
cr
;

\ 100*base values
161 constant bubblebase
175 constant permbase
183 constant queenbase
239 constant towersbase
192 constant quickbase
146 constant intmmbase
250 constant treebase
50 constant puzzlebase






." Nonfloating point composite is "
." Nonfloating geometric mean is "


\ Sun 3/280 (cc -O3) results
\     Perm    0.393
\   Towers    0.512
\   Queens    0.173
\    Intmm    0.260
\   Puzzle    1.120
\    Quick    0.190
\   Bubble    0.225
\     Tree    0.662
\       Mm    5.035
\      FFT    9.482

\ Nonfloating point composite is   5.549
\ Floating point composite is     62.350
\ Nonfloating geometric mean is   0.361
\ Geometric mean is               0.651



